#!/usr/local/bin/ruby -w

require 'ncurses'
require 'ctype'

#module Ncurses
	#def curs_set(bool)
	#end
#end

class VINode < Array
	# What is the text of this node?
	attr_reader :text

	@@id = 0
	def initialize(text)
		@text = text
		@id = @@id
		@@id += 1
	end
end


def fmt_tokenize(s)
	if s[0] == 10
		return s[1..-1], s[0..0]
	else
		return s.gsub(/\A([ \t]+|\S+)/m, ''), $1
	end
end
def fmt(str, width)
	# Special case when we have an empty string
	return [''] if str == ''
	# Strip off control characters except newline, replacing a run of them
	# with a single space.
	# XXX Assumes ASCII/ISO-8859, and does not accept the Windows messing
	# with these.
	str.gsub!(/[\x0-\x9\xb-\x1f\x80-\x9f]+/, ' ')
	array = []
	outstr = ""
	space = ""
	while str != ''
		str, token = fmt_tokenize(str)
		if token == "\n"
			array.push(outstr)
			outstr = ""
			space = ""
		elsif token =~ /^\s/
			if outstr.length + token.length >= width
				space = ""
				array.push(outstr)
				outstr = ""
			else
				space = token
			end
		else
			if outstr.length + space.length + token.length <= width
				outstr += space + token
			else
				if token.length < width
					array.push(outstr)
					outstr = token
				else
					# Handle special case for single words
					# that are too long to fit in a line.
					array.push(outstr) if outstr != ""
					while token.length > width
						token.gsub!(/^(.{#{width}})/, '')
						array.push($1)
					end
					outstr = token
				end
			end
			# Never care about space when we have just done a word
			space = ""
		end
	end
	array.push(outstr) if outstr
	return array
end

class ScreenContext
	attr_reader :indent
	def initialize(indent)
		@indent = indent
	end
	def nlines(text)

	end
end

PREF_Indent=4		# Indent preference

#
# Function to render the a set of nodes to the screen.
# Takes as a parameter the present path of nodes to render, the X and Y
# position of the cursor on the screen, and the offset into the line the cursor
# presently is on.
#
def render(
	rootnode,	# Root of entire tree
	nodepath, 	# Path to root node from active node
	x, y,		# Position of the cursor on the screen
	ox, oy)		# Offset X and Y inside the present node
		cols = Ncurses.COLS	# XXX
		indent = nodepath.length * PREF_Indent
		# Wrap back to the start of the line if we get too long lines
		# (ie, start to the right of the screen).  Otherwise, use
		# indent as it is.
		width = cols - (indent % cols)

=begin
		mainout = fmt(nodepath[-1][0].text, width)
		raise "Bad fmt result" unless mainout.length > 0
		# Constrain ox/oy to be inside the node.
		# xxx Should this be assertions? Let's see later.
		if (oy < mainout.length)
			ox = 0 if oy > mainout[ox].length
		else
			oy = mainout.length - 1
			ox = 0
		end
=end
		maxy = cols - 1 - y
		nfc = NodeFormatContext.new(rootnode, nodepath, ox, oy, Ncurses.COLS)
		# Print below the present position
		0.upto(maxy) { |dy|
			node, line = nfc.getline(dy)
			Ncurses.attron(Ncurses::WA_BOLD) if node && node.length > 0
			Ncurses.standout if node == nfc.homenode
			Ncurses.mvaddstr(y+dy, 0, line)
			Ncurses.standend if node == nfc.homenode
			Ncurses.attroff(Ncurses::WA_BOLD) if node && node.length > 0
		}
		y.downto(0) { |dy|
			node, line = nfc.getline(dy - y)
			Ncurses.attron(Ncurses::WA_BOLD) if node && node.length > 0
			Ncurses.standout if node == nfc.homenode
			Ncurses.mvaddstr(dy, 0, line)
			Ncurses.standend if node == nfc.homenode
			Ncurses.attroff(Ncurses::WA_BOLD) if node && node.length > 0
		}
end

def getnode(rootnode, nodepath)
	nodepath = nodepath.dup
	node = rootnode
	while offset = nodepath.shift
		node = node[offset]
	end
	return node
end
#
# This object is capable of formatting text from a node stack, and give back a
# set of lines.
# This is FAR from an optimal way of formatting everything, but it is simple to
# implement, and I do not believe I'll lack CPU power for this app anyway.
#
# XXX It would be nice to have the object cache the pre-formatted data, though.
#
class NodeFormatContext
   public
	attr_reader :homenode
	def initialize(
		rootnode,	# The root of the entire tree (text not displayed)
		nodepath, 	# Path to root node from active node
		ox, oy,		# Offset X and Y inside the present node
		width)		# Width to assume formatting is happening in
		@rootnode = rootnode
		@nodepath = nodepath
		@ox, @oy = ox, oy
		@width = width
		@homenode = getnode(nodepath).text
		@homefmt = fmt(@homenode, width - PREF_Indent * nodepath.length)
	end
    private
	def getnode(nodepath)
		nodepath = nodepath.dup
		node = @rootnode
		while offset = nodepath.shift
			node = node[offset]
		end
		return node
	end
    public
	def getline(
		y)		# offset from origin, in lines
		orig_y = y

		# Special case lines that are inside our "home node" - the main
		# node that is currently active.
		if (y + @oy >= 0 && y + @oy < @homefmt.length)
			return @homenode, (" " * @nodepath.length * PREF_Indent) + @homefmt[@oy + y]
		end
		return nil, "" unless @nodepath.length > 1

		nodepath = @nodepath.dup
		#nodepath.pop
		if y < 0
			#
			# Iterate backwards through stack
			#

			# Lose offset
			y += @oy
			while nodepath.length > 0
				if nodepath[-1] == -1
					nodepath.pop
					next
				end
				currentnode = getnode(nodepath)
				currenttext = fmt(currentnode.text, @width - PREF_Indent * nodepath.length)
				if (currenttext.length + y > 0)
					return currentnode,
						(" " * nodepath.length * PREF_Indent) + currenttext[y - 1]
				else
					y += currenttext.length
				end
				nodepath[-1] = nodepath[-1] - 1
			end
		else
			#
			# Iterate forwards through stack
			#

			# Lose offset
			y += @homefmt.length - @oy
			# Get rid of present node
			#nodepath[-1] += 1
			# Display
			while nodepath.length > 0
				nodepath2 = nodepath.dup
				nodepath2.pop
				if nodepath[-1] == getnode(nodepath2).length
					nodepath.pop
					next
				end
				currentnode = getnode(nodepath)
				currenttext = fmt(currentnode.text, @width - PREF_Indent * nodepath.length)
				if (y < currenttext.length)
					return currentnode, 
						(" " * nodepath.length * PREF_Indent) + currenttext[y]
				else
					y -= currenttext.length
				end
				nodepath[-1] = nodepath[-1] + 1
			end
		end
		return nil, ""
	end
end


# root
root = VINode.new("rootnode")
# root[0]
root.push(VINode.new("This is a test node.  It only contains a bunch of text, \
and nothing really interesting.  However, it helps test the display code, \
and that can probably be really useful.  It does not have any children."))
# root[1]
root.push(VINode.new("This is a parent node."))
# root[1][0]
root[1].push(VINode.new("This is a child node, with children."))
# root[1][0][0]
root[1][0].push(VINode.new("This is a grandchild."))
# root[1][1]
root[1].push(VINode.new("Another child, this one without children."))
root[1].push(VINode.new("[1][0][2]"))
# root[2]
root.push(VINode.new("And another from the root."))
# root[2][0]
root[2].   push(VINode.new("New child"))
root[2][0].push(VINode.new("[2][0][0]"))
root[2].   push(VINode.new("[2][1]"))

# The path from the lowest displayed node to the root. It is necessary to track
# the path because the nodes themselves can be multiple places in the tree.
#nodepath = [root, root[1], root[1][1]]
nodepath = [root, root[0]]
		# Node,		# Retrieved from position in parent
nodepath2 = [	[root,		0],
	    	[root[1],	1],
	    	[root[1][1],	1]]

def recprint(prefix, array)
	array.each_index { |i|
		print "#{prefix}[#{i}]\n"
		if array[i].kind_of? Array
			recprint("#{prefix}[#{i}]", array[i])
		end
	}
end
$stderr.sync = true

#recprint("", root)
#exit 0
# Available indexes:
# [0]
# [1]
# [1][0]
# [1][0][0]
# [1][1]
# [2]

def prevnode(rootnode, nodepath)
	nodepath = nodepath.dup
	if nodepath.length == 0
		return [0]
	elsif nodepath[-1] > 0
		nodepath[-1] = nodepath[-1] - 1
	else
		while nodepath.length > 0 && nodepath[-1] == 0
			nodepath.pop
		end
		return [0] if nodepath.length == 0
	end
	return nodepath
end

def nextnode(rootnode, nodepath)
end

class NodePath
end
	

char = nil
Ncurses.initscr()
begin
	Ncurses.nonl		# Do not force newline; gives better curses
				# performance
	Ncurses.curs_set(0)	# Disable cursor for now
	Ncurses.cbreak()
	Ncurses.noecho();
	Ncurses.intrflush(Ncurses.stdscr, FALSE);
	Ncurses.keypad(Ncurses.stdscr, TRUE);

	
	nodepath = [1, 0, 0]
	while true
		render(root, nodepath, 30, 20, 0, 0)
		char = Ncurses.getch
		case char
		when 107	# k
			nodepath = prevnode(root, nodepath)
		when 113	# q
			break
		end
	end
		
ensure
	Ncurses.endwin
end
print "Char: #{char}\n"
